unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ComCtrls, ToolWin, ExtCtrls, ImgList, Globals, Render;

type
  TMainForm = class(TForm)
    MainMenu: TMainMenu;
    MGame: TMenuItem;
    ToolBar: TToolBar;
    Status: TStatusBar;
    ThePanel: TPanel;
    TheBox: TPaintBox;
    Images: TImageList;
    BNew: TToolButton;
    BPlay: TToolButton;
    Separator1: TToolButton;
    BUndo: TToolButton;
    BRedo: TToolButton;
    MNew: TMenuItem;
    BRasterSave: TToolButton;
    MRasterSave: TMenuItem;
    N1: TMenuItem;
    MQuit: TMenuItem;
    MDebug: TMenuItem;
    MShowRF1: TMenuItem;
    MShowRF2: TMenuItem;
    MShowField: TMenuItem;
    N2: TMenuItem;
    MShowPossibleOwner: TMenuItem;
    BStop: TToolButton;
    BPause: TToolButton;
    BContinue: TToolButton;
    Hilfe1: TMenuItem;
    N3: TMenuItem;
    About1: TMenuItem;
    Spielanleitung1: TMenuItem;
    MView: TMenuItem;
    MZoom50: TMenuItem;
    MZoom100: TMenuItem;
    MZoom200: TMenuItem;
    MZoomIn: TMenuItem;
    MZoomOut: TMenuItem;
    N4: TMenuItem;
    MHint: TMenuItem;
    N5: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure StatusDrawPanel(StatusBar: TStatusBar; Panel: TStatusPanel;
      const Rect: TRect);
    procedure MNewClick(Sender: TObject);
    procedure BNewClick(Sender: TObject);
    procedure TheBoxPaint(Sender: TObject);
    procedure FieldDesignMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FieldDesignMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure BPlayClick(Sender: TObject);
    procedure MRasterSaveClick(Sender: TObject);
    procedure MGameClick(Sender: TObject);
    procedure GameMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MQuitClick(Sender: TObject);
    procedure MShowFieldClick(Sender: TObject);
    procedure MShowRF1Click(Sender: TObject);
    procedure MShowRF2Click(Sender: TObject);
    procedure MDebugClick(Sender: TObject);
    procedure MShowPossibleOwnerClick(Sender: TObject);
    procedure BUndoClick(Sender: TObject);
    procedure BRedoClick(Sender: TObject);
    procedure BStopClick(Sender: TObject);
    procedure BPauseClick(Sender: TObject);
    procedure BContinueClick(Sender: TObject);
    procedure Spielanleitung1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure MZoomClick(Sender: TObject);
    procedure MZoomRel(Sender: TObject);
    procedure MHintClick(Sender: TObject);
  private
    { Private declarations }
    Locked: boolean;
    UIState: TUIState;
    Renderer: TFieldRenderer;
    MarkX, MarkY: integer;
    MarginWidth, MarginHeight: integer;
    procedure SetUIState(NewState: TUIState);
    procedure UpdateWindowSize;
  public
    { Public declarations }
    procedure NotifyUpdate;
    procedure Lock;
    procedure Unlock;
  end;

var
  MainForm: TMainForm;

implementation

uses
  NewGame, RasterSave, Game, RasterField, Fields, Player, Rules, About;

{$R *.dfm}

procedure TMainForm.FormCreate(Sender: TObject);
begin
//AllocConsole;
  Randomize;
  MarginWidth := Width - TheBox.Width;
  MarginHeight := Height - TheBox.Height;
  TheGame := TGame.Create;
  Renderer := TFieldRenderer.Create;

  Locked := false;
  UIState := uisStartup;
  Status.SimpleText := Caption + ' Build ' + BuildID + ' (' + Creator + ')';
  TheBox.ControlStyle := TheBox.ControlStyle + [csOpaque];
  MDebug.Visible := (ParamStr(1) = '-debug');
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Renderer.Free;
  if Assigned(TheGame) then TheGame.Free;
end;

procedure TMainForm.SetUIState(NewState:TUIState);
begin
  case NewState of
    uisDesign:begin
                TheBox.OnMouseDown := FieldDesignMouseDown;
                TheBox.OnMouseMove := FieldDesignMouseMove;
                TheBox.OnMouseUp   := nil;
                Status.SimplePanel := true;
                Status.SimpleText := 'Feld erstellen = linke Maustaste; '+
                                     'Feld entfernen = rechte Maustaste';
                Renderer.ShowField := false;
                BNew.Visible := false;
                BPlay.Visible := true;
                BStop.Visible := false;
                BPause.Visible := false;
                BContinue.Visible := false;
                BUndo.Visible := false;
                BRedo.Visible := false;
                BRasterSave.Visible := true;
              end;
    uisGame:begin
              TheBox.OnMouseDown := GameMouseDown;
              TheBox.OnMouseMove := nil;
              TheBox.OnMouseUp   := nil;
              Status.SimplePanel := false;
              Renderer.ShowField := true;
              BNew.Visible := true;
              BPlay.Visible := false;
              BUndo.Visible := true;
              BRedo.Visible := true;
              BRasterSave.Visible := false;
            end;
  end;
  TheBox.OnPaint := TheBoxPaint;
  MarkX := -1;
  MarkY := -1;
  UIState := NewState;
end;

procedure TMainForm.StatusDrawPanel(StatusBar: TStatusBar;
  Panel: TStatusPanel; const Rect: TRect);
begin
  case Panel.ID of
    0:with StatusBar.Canvas do begin
        Brush.Color := Renderer.PlayerColors[TheGame.CurrentPlayer];
        FillRect(Rect);
      end;
    1:with StatusBar.Canvas do begin
        Font.Style := [fsBold];
        TextOut(Rect.Left+2, Rect.Top+1,
                TheGame.Player[TheGame.CurrentPlayer].Name)
      end;
  end;
end;

procedure TMainForm.MNewClick(Sender: TObject);
begin
  NewGameForm.PColor1.Color := Renderer.PlayerColors[0];
  NewGameForm.PColor2.Color := Renderer.PlayerColors[1];
  if NewGameForm.ShowModal <> mrOK then exit;
  if Assigned(TheGame.Raster)  then TheGame.Raster.Free;
  if Assigned(TheGame.Player[0]) then TheGame.Player[0].Free;
  if Assigned(TheGame.Player[1]) then TheGame.Player[1].Free;
  if NewGameForm.CCustomField.Checked then begin
    TheGame.Raster := TRaster.Create(NewGameForm.CWidth .Position,
                                     NewGameForm.CHeight.Position);
    SetUIState(uisDesign);
  end else begin
    TheGame.Raster := TRaster.Create(NewGameForm.CSelectField.Items[
                                     NewGameForm.CSelectField.ItemIndex]);
    if Assigned(TheGame.Field) then TheGame.Field.Free;
    TheGame.Field := TField.Create(TheGame.Raster);
    SetUIState(uisGame);
  end;
  Renderer.PlayerColors[0] := NewGameForm.PColor1.Color;
  Renderer.PlayerColors[1] := NewGameForm.PColor2.Color;
  TheGame.Player[0] := TPlayer.Create(0, TheGame.Raster,
                       NewGameForm.EName1.Text, NewGameForm.CHuman1.Checked,
                       NewGameForm.CAILevel1.Position);
  TheGame.Player[1] := TPlayer.Create(1, TheGame.Raster,
                       NewGameForm.EName2.Text, NewGameForm.CHuman2.Checked,
                       NewGameForm.CAILevel2.Position);
  UpdateWindowSize;
  if UIState = uisGame then TheGame.Restart;
end;

procedure TMainForm.BNewClick(Sender: TObject);
begin
  if UIState = uisStartup then begin
    MNewClick(nil);
    exit;
  end;
  TheGame.Restart;
end;

procedure TMainForm.TheBoxPaint(Sender: TObject);
begin
  Renderer.RenderField(TheGame.Field, TheGame.Raster, TheBox.Canvas);
end;

procedure TMainForm.FieldDesignMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  MarkX := -1;
  MarkY := -1;
  case Button of
    mbLeft: FieldDesignMouseMove(Sender, Shift + [ssLeft], X, Y);
    mbRight: FieldDesignMouseMove(Sender, Shift + [ssRight], X, Y);
  end;
end;

procedure TMainForm.FieldDesignMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
  if not(ssLeft IN Shift) AND not(ssRight IN Shift) then exit;
  X := X DIV Renderer.FieldSize;
  Y := Y DIV Renderer.FieldSize;
  if (X = MarkX) AND (Y = MarkY) then exit;
  if (X < 1) OR (X >= TheGame.Raster.Width  - 1) OR
     (Y < 1) OR (Y >= TheGame.Raster.Height - 1) then exit;
  TheGame.Raster[X, Y] := (ssLeft IN Shift);
  MarkX := X; MarkY := Y;
  TheBox.Repaint;
end;

procedure TMainForm.BPlayClick(Sender: TObject);
begin
  SetUIState(uisGame);
  TheGame.Restart;
end;

procedure TMainForm.MGameClick(Sender: TObject);
begin
  MRasterSave.Enabled := (UIState <> uisStartup);
end;

procedure TMainForm.MRasterSaveClick(Sender: TObject);
begin
  if RasterSaveForm.ShowModal <> mrOK then exit;
  if not TheGame.Raster.Save(RasterSaveForm.CSaveName.Text) then
    MessageDlg('Konnte Spielfeld nicht speichern!', mtError, [mbOK], 0);
end;

procedure TMainForm.NotifyUpdate;
begin
  Status.Repaint;
  TheBox.Repaint;
end;

procedure TMainForm.GameMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var MX,MY:integer;
    DX,DY:integer;
    Move:TMove;
begin
  if not Assigned(TheGame) then exit;
  if not TheGame.Player[TheGame.CurrentPlayer].IsHuman then exit;
  // Punktkoordinaten bestimmen
  with Renderer do begin
    X := (X - FieldSize DIV 2) DIV FieldSize;
    Y := (Y - FieldSize DIV 2) DIV FieldSize;
  end;
  // ungltige Koordinaten ablehnen
  if (X < 0) OR (X >= TheGame.Field.Width) OR
     (Y < 0) OR (Y >= TheGame.Field.Height) then exit;

  // vvvvvv DEBUG
  if Button<>mbLeft then begin
    {if TheGame.Field.MayBeOccupiedBy(TheGame.CurrentPlayer,X,Y)
      then Status.Panels[2].Text:='ja'
      else Status.Panels[2].Text:='n';}
    exit;
  end;
  // ^^^^^^ DEBUG

  // Noch kein Feld markiert ==> ggf. Markierung setzen
  if (MarkX < 0) OR (MarkY < 0) then begin
    if TheGame.Field[X, Y].Owner <> OWNER_OOF then begin
      MarkX := X;
      MarkY := Y;
      Renderer.HighlightNode(X, Y, TheBox.Canvas);
    end;
    exit;
  end;
  // Feld markiert ==> Vorarbeiten durchfhren
  MX := MarkX;
  MY := MarkY;
  MarkX := -1;
  MarkY := -1;
  // Klick auerhalb des Spielfelds ==> Abbruch
  if TheGame.Field[X, Y].Owner = OWNER_OOF then begin
    TheBox.Repaint;
    exit;
  end;
  // sehr elegante "Formel" zum Ermitteln des mittleren Feldes und der Richtung
  DX := MX - X;
  DY := MY - Y;
  if (DX AND 1 <> 0) OR (DY AND 1 <> 0) then begin
    TheBox.Repaint;
    exit;
  end;
  case (abs(DX) + abs(DY)) of
    2:if DX <> 0 then Move.Dir := DIR_H
                 else Move.Dir := DIR_V;
    4:if (DX * DY) < 0 then Move.Dir := DIR_D
                       else Move.Dir := DIR_A;
  else
    TheBox.Repaint;
    exit;
  end;
  // Zugstruktur basteln
  Move.X := (MX + X) DIV 2;
  Move.Y := (MY + Y) DIV 2;
  Move.Player := TheGame.CurrentPlayer;
  // Zug durchfhren
  if not TheGame.CommitMove(Move) then TheBox.Repaint;
end;

procedure TMainForm.MQuitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.MShowFieldClick(Sender: TObject);
begin
  Renderer.ConnectLines := true;
  Renderer.RenderField(TheGame.Field, TheGame.Raster, TheBox.Canvas);
  Status.Panels[2].Text := '';
end;

procedure TMainForm.MShowRF1Click(Sender: TObject);
begin
  Renderer.ConnectLines := false;
  Renderer.RenderField(TheGame.Player[0].RemainingField,
                       TheGame.Raster, TheBox.Canvas);
  Status.Panels[2].Text := IntToStr(TheGame.Player[0].RemainingMoves)
                           + ' remaining moves.';
end;

procedure TMainForm.MShowRF2Click(Sender: TObject);
begin
  Renderer.ConnectLines := false;
  Renderer.RenderField(TheGame.Player[1].RemainingField,
                       TheGame.Raster, TheBox.Canvas);
  Status.Panels[2].Text := IntToStr(TheGame.Player[1].RemainingMoves)
                           + ' remaining moves.';
end;

procedure TMainForm.MDebugClick(Sender: TObject);
begin
  MShowPossibleOwner.Checked := Renderer.ShowPossibleOwner;
end;

procedure TMainForm.MShowPossibleOwnerClick(Sender: TObject);
begin
  Renderer.ShowPossibleOwner := not(Renderer.ShowPossibleOwner);
  TheBox.Repaint;
end;

procedure TMainForm.BUndoClick(Sender: TObject);
begin
  TheGame.Xdo(xdoUndo);
end;

procedure TMainForm.BRedoClick(Sender: TObject);
begin
  TheGame.Xdo(xdoRedo);
end;

procedure TMainForm.Lock;
begin
  Locked := true;
  BNew.Visible := false;
  BStop.Visible := true;
  BUndo.Enabled := false;
  BRedo.Enabled := false;
  MGame.Enabled := false;
  MView.Enabled := false;
  MDebug.Enabled := false;
  Application.ProcessMessages;
end;

procedure TMainForm.Unlock;
begin
  Locked := false;
  BNew.Visible := true;
  BStop.Visible := false;
  BUndo.Enabled := true;
  BRedo.Enabled := true;
  MGame.Enabled := true;
  MView.Enabled := true;
  MDebug.Enabled := true;
  Invalidate;
end;

procedure TMainForm.BStopClick(Sender: TObject);
begin
  AbortMove := true;
end;

procedure TMainForm.BPauseClick(Sender: TObject);
begin
  BPause.Visible := false;
  BContinue.Visible := true;
  TheGame.Paused := true;
end;

procedure TMainForm.BContinueClick(Sender: TObject);
begin
  BPause.Visible := true;
  BContinue.Visible := false;
  TheGame.Paused := false;
  if not Locked then TheGame.NextPlayer;
end;

procedure TMainForm.Spielanleitung1Click(Sender: TObject);
begin
  RuleForm.ShowModal;
end;

procedure TMainForm.About1Click(Sender: TObject);
begin
  AboutForm.ShowModal;
end;

procedure TMainForm.MZoomClick(Sender: TObject);
begin
  with Sender as TMenuItem do
    Renderer.Zoom := Tag;
  UpdateWindowSize;
end;

procedure TMainForm.MZoomRel(Sender: TObject);
begin
  with Sender as TMenuItem do
    Renderer.Zoom := Renderer.Zoom + Tag;
  UpdateWindowSize;
end;

procedure TMainForm.UpdateWindowSize;
begin
  if not Assigned(TheGame) then exit;
  Width  := TheGame.Raster.Width  * Renderer.FieldSize + MarginWidth;
  Height := TheGame.Raster.Height * Renderer.FieldSize + MarginHeight;
  TheBox.Repaint;
end;

procedure TMainForm.MHintClick(Sender: TObject);
var Move:  TMove;
    BX,BY: integer;
    LLL:   integer;
begin
  if not Assigned(TheGame) then exit;
  with TheGame,Player[CurrentPlayer] do begin
    if not IsHuman then exit;
    Move := DetermineRandomBestDefensiveMove;
  end;

  with TheBox.Canvas,Renderer do begin
    Pen.Width := 1;
    Pen.Style := psDot;
    Pen.Color := Renderer.PlayerColors[TheGame.CurrentPlayer];
    BX  := FieldSize * (Move.X + 1);
    BY  := FieldSize * (Move.Y + 1);
    LLL := FieldSize;
    case Move.Dir of
      DIR_H: begin MoveTo(BX - LLL, BY);       LineTo(BX + LLL, BY);       end;
      DIR_V: begin MoveTo(BX,       BY - LLL); LineTo(BX,       BY + LLL); end;
      DIR_D: begin MoveTo(BX - LLL, BY + LLL); LineTo(BX + LLL, BY - LLL); end;
      DIR_A: begin MoveTo(BX - LLL, BY - LLL); LineTo(BX + LLL, BY + LLL); end;
    end;
  end;
end;

end.
